package tech.loedige;

public class KeyStreamGenerator_ChaCha20  extends SynchronousKeyStreamGenerator {
        // Zustand des Schlüsselstromgenerators
        // definiert durch (s0, s, c)
        private final int[] s0 = new int[16];
        private final int[] s = new int[16];
        private long c;
        /**
         * Initialisierung des Schlüsselstromgenerators (Funktion s_0)
         *
         * @param key Schlüssel zu Initialisierung des
         * Schlüsselstromgenerators (Bytearray der Länge 32)
         * @param nonce Nonce-Wert zu Initialisierung des
         * Schlüsselstromgenerators (Bytearray der Länge 12)
         */
        public void initState(byte[] key, byte[] nonce) {
            initState(key, nonce, 0);
        }
        /**
         * Initialisierung des Schlüsselstromgenerators (Funktion s_0 mit zusätzlicher Initialisierung der
         * Zählervariablen in der ersten Komponente der Zustandsvariablen s0[12])
         * @param key Schlüssel zu Initialisierung des
         * Schlüsselstromgenerators (Bytearray der Länge 32)
         * @param nonce Nonce-Wert zu Initialisierung des
         * Schlüsselstromgenerators (Bytearray der Länge 12)
         * @param blockCtr Initialwert des Zählerwerts in s0[12]
         */
        public void initState(byte[] key, byte[] nonce, int blockCtr) {
            if ((key.length != 32) || (nonce.length != 12)) {
                throw new IllegalArgumentException("Invalid length of key or nonce!");
            }
            s0[0] = 0x61707865;
            s0[1] = 0x3320646e;
            s0[2] = 0x79622d32;
            s0[3] = 0x6b206574;
            // Transformation der Schlüsselbytes in int-Werte
            // s0[4},...,s0[11] (Die Bytes werden in ’little-endian order’ eingelesen.)
            for (int i = 0; i < 8; ++i) {
                s0[i + 4] = (0xff & key[4 * i]);
                s0[i + 4] |= (0xff & key[4 * i + 1]) << 8;
                s0[i + 4] |= (0xff & key[4 * i + 2]) << 16;
                s0[i + 4] |= (0xff & key[4 * i + 3]) << 24;
                }
            s0[12] = blockCtr;

            // Transformation der Nonce-Bytes in int-Werte
            // s0[13},s0[14],s0[15]
            for (int i = 0; i < 3; ++i) {
                s0[i + 13] = (0xff & nonce[4 * i]);
                s0[i + 13] |= (0xff & nonce[4 * i + 1]) << 8;
                s0[i + 13] |= (0xff & nonce[4 * i + 2]) << 16;
                s0[i + 13] |= (0xff & nonce[4 * i + 3]) << 24;
                }
            c = 0;

            // Aufruf von updateStateBlocks() (entspricht der
            // Funktion s) wird durch Aufruf von
            // getNextKeyStreamByte() (entspricht der Funktion S)
            // angestossen.
            }


            /**
             * Berechnung des nächsten Schlüsselstrombytes.
             * (Entspricht eines einmaligen Aufrufs der Funktionen s
             * und S.)
             *
             * @return Nächstes Schlüsselstrombyte
             */
            @Override
            public byte getNextKeyStreamByte() {
                if ((c & 0x3f) == 0) { // (c mod 64) == 0
                    if (c == 0x4000000000L) { // 2^38
                        throw new RuntimeException("Keystream exhausted!");
                    }
                    updateStateBlocks();
                }
                int stateWord = s[(int) ((c >> 2) & 0x0f)];
                byte keyStreamByte= (byte) (stateWord >> (int) 8 * (c & 0x03));
                ++c;
                return keyStreamByte;
            }

            /**
             * Update (Funktion s) der Block-Komponenten des Zustands100
             * (für (c mod 64) == 0)
             */
            private void updateStateBlocks() {
                System.arraycopy(s0, 0, s, 0, 16);
                // 10-fache Anwendung der Inner-Block_Funktion f_{ib}
                for (int i = 0; i < 10; ++i) {
                    quaterround(0, 4, 8, 12);
                    quaterround(1, 5, 9, 13);
                    quaterround(2, 6, 10, 14);
                    quaterround(3, 7, 11, 15);
                    quaterround(0, 5, 10, 15);
                    quaterround(1, 6, 11, 12);
                    quaterround(2, 7, 8, 13);
                    quaterround(3, 4, 9, 14);
                }
                for (int i = 0; i < 16; ++i) {
                    s[i] += s0[i];
                }
                s0[12] += 1; // Inkrementierung des Block-Zählers
                }

                /**
                 * Quater-Round-Funktion (Funktion q)
                 */
                private void quaterround(int i1, int i2, int i3, int i4) {
                    int a = s[i1];
                    int b = s[i2];
                    int c = s[i3];
                    int d = s[i4];
                    a += b;
                    d ^= a;
                    d = (d << 16) | (d >>> 16);
                    c += d;
                    b ^= c;
                    b = (b << 12) | (b >>> 20);
                    a += b;
                    d ^= a;
                    d = (d << 8) | (d >>> 24);
                    c += d;
                    b ^= c;
                    b = (b << 7) | (b >>> 25);
                    s[i1] = a;
                    s[i2] = b;
                    s[i3] = c;
                    s[i4] = d;
                    }

            }

